// -*- Mode: Go; indent-tabs-mode: t -*-

/*
 * Copyright (C) 2019 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package sanity

import (
	"fmt"

	"github.com/snapcore/snapd/sandbox/cgroup"
)

func init() {
	checks = append(checks, checkCgroup)
}

func checkCgroup() error {
	v, err := cgroup.Version()
	if err != nil {
		return fmt.Errorf("snapd could not probe cgroup version: %v", err)
	}
	if v == cgroup.Unknown {
		return fmt.Errorf("snapd could not determine cgroup version")
	}

	return nil
}
