/*
This package provides bindings to the generic input event interface in
Linux. The evdev interface serves the purpose of passing events
generated in the kernel directly to userspace through character
devices that are typically located in /dev/input/.

Please refer to the godoc examples and the bin/evtest example program.

*/
package evdev
